#
/*
 *	    M U L T I - D E V I C E   Q U E U E I N G   S Y S T E M
 *
 *			   Douglas P. Kingston III
 *
 *			     Batch/Interactive
 *		     Network Request Receiver  (NETRECV)
 *
 *	________________________________________________________________
 *
 *				Revision History
 *				================
 */

#include <stdio.h>
#include <signal.h>
#include "queue.h"

extern	char *Queue2;
extern	char *Tmpqdir;

char	*index();
long	atol();
struct	request *gethdr();

char	request[LINESIZ];
char	site[LINESIZ];
char	queue[LINESIZ];
char	user[LINESIZ];
char	file[LINESIZ];
char	*host = 0;
char	*accessfile = 0;
char	*sitename = "The Mystery Host";
int	interactive = 0;
int	oldformat = 0;		/* if set, allow old format */
int	debug = 0;
char	inbuf[BUFSIZ];
char	missingfmt[] = "Missing %s argument";
char	dupfmt[] = "Duplicate %s argument given";
char	newlinefmt[] = "Illegal newline in %s argument";

struct paramtab paramtab[] = {
	"accessfile",	STRPVAR,	&accessfile,		NULL,
	"hostname",	STRPVAR,	&sitename,		NULL,
	0,		0,		0,			0
};

/*
 *		M A I N  ( )
 */
main (argc, argv)
int	argc;
char	**argv;
{
	int	alarmtrap();
	char	buf[LINESIZ];
	long	filesize;
	char	*arg;
	int	i;
	FILE	*fp;

	umask (077);
	signal (SIGALRM, alarmtrap);
	rdconf( paramtab );
	if (chdir (Tmpqdir) < 0) {
		netclose ("512 Cannot chdir to Tmpqdir");
		exit (1);
	}

	for (i = 1; i < argc; i++) {
		if (*argv[i] != '-')
			break;
		arg = argv[i];
		arg++;
		while (*arg) {
			switch (*arg) {
			case 'd':
				debug++;
				break;

			case 'h':	/* Name of requesting host follows */
				if (argv[i+1] == 0) {
					err (missingfmt, "host");
					exit (99);
				} else
					i++;
				if (host != 0) {
					err (dupfmt, "host");
					exit (98);
				}
				host = argv[i];
				break;

			case 'i':
				interactive++;
				break;

			case 'o':
				oldformat++;
				break;

			default:
				err ("NETRECV: Unknown flag %c", *arg);
				exit (97);
			}
			arg++;
		}
	}

if (debug) fputs ("NETRECV started\n", stderr);
	netreply ("100 %s MDQS Service; I think you are %s",
		  sitename, host);
	setbuf (stdin, inbuf);
	getline (buf, LINESIZ);
	if (!lexnequ (buf, "Site: ", 6)) {
		netclose ("500 Missing site name");
		exit (1);
	}
	strcpy (site, &buf[6]);
	netreply ("101 Doesn't this seem redundant!");
	
if (debug) fputs ("got site:\n", stderr);
	getline (buf, LINESIZ);
	if (!lexnequ (buf, "User: ", 6)) {
if (oldformat)		/** KLUDGE **/
  goto next_old_field;	/** KLUDGE **/
		netclose ("509 Missing user name");
		exit (1);
	}
	strcpy (user, &buf[6]);
	netreply ("102 I'll take your word for it.");

if (debug) fputs ("got user:\n", stderr);
	getline (buf, LINESIZ);
	if (!lexnequ (buf, "Queue: ", 7)) {
		netclose ("507 Missing queue name");
		exit (1);
	}
	strcpy (queue, &buf[7]);
	if (host && ! netaccess(host, queue)) {
		err ("Access denied to host %s, user %s, queue %s",
		     host, user, queue);
		netclose ("508 Not authorized to submit to this queue");
		exit (1);
	}
	netreply ("103 Queue name accepted");

if (debug) fputs ("got queue:\n", stderr);
	getline (buf, LINESIZ);
next_old_field:
	if (!lexnequ (buf, "Request: ", 9)) {
		netclose ("501 Missing request name");
		exit (1);
	}
	strcpy (request, &buf[9]);
	netreply ("104 Request name accepted");

if (debug) fputs ("got request:\n", stderr);
	for (;;) {
		getline (buf, LINESIZ);
		if (lexnequ (buf, "Done", 4))
			doqueue();
		else if (!lexnequ (buf, "File: ", 6)) {
			netclose ("502 Missing file name");
			exit (1);
		}
		strcpy (file, &buf[6]);
		if (index (file, '/') != 0) {
			netclose ("503 File name not permitted");
			exit (1);
		}
		if ((fp = fopen(file, "w")) == NULL) {
			netclose ("400 Cannot open file");
			exit (2);
		}
		netreply ("105 Filename OK");

		getline (buf, LINESIZ);
		if (!lexnequ (buf, "Size: ", 6)) {
			netclose ("505 Missing file size");
			exit (1);
		}
		filesize = atol (&buf[6]);
		if (filesize < 0L || (filesize == 0L && buf[6] != '0')) {
			netclose ("506 Bad file size");
			exit (1);
		}
		netreply ("106 Size OK, fire away!");
		
		suckup (filesize, fp);
		fflush (fp);
		if (ferror (fp)) {
			netclose ("401 Error writing queue file");
			exit (2);
		}
		fclose (fp);
		netreply ("107 Got your data, what's next?");
	}
}

/***** On 4.2, this should be using select() *************/
getline (buf, bufsiz)
char *buf;
register int bufsiz;
{
	register char c;
	register int i;

	i = 0;
	alarm (30);
	while ((c = getchar()) != '\n') {
		switch (c) {
		case EOF:
			alarm (0);
			netclose ("500 Premature EOF, Sorry Charlie.");
			exit (1);
		default:
			buf[i++] = c;
			break;
		}
		if (i >= bufsiz) {
			alarm (0);
			netclose ("501 Filename too long.");
			exit (1);
		}
	}
	alarm (0);
	buf[i] = 0;
if (debug) fprintf(stderr, "%s\n", buf);
}

suckup (cnt, fp)
long	cnt;
FILE	*fp;
{
	char	buf[BUFSIZ];
	int	i;
	
	while (cnt > 0L) {
		if (interactive)
			alarm (30);
		i = fread (buf, 1, (cnt > BUFSIZ) ? BUFSIZ : (int) cnt, stdin);
		if (interactive)
			alarm (0);
		if (i == 0 || ferror(stdin)) {
			netclose ("450 Premature EOF");
			exit(2);
		} else {
			if (fwrite (buf, 1, i, fp) != i || ferror(fp)) {
				netclose ("451 Error on queue file");
				exit (2);
			}
			cnt -= i;
		}
	}
}

doqueue ()
{
	FILE	*rfp;
	struct	request r;
	int	status;
	int	pfd[2];
	
	if ((rfp = fopen (file, "r+")) == NULL) {
		netclose ("404 Cannot reopen request file");
		exit (2);
	}
	if (gethdr (rfp, &r) == (struct request *) NULL) {
		netclose ("405 Cannot read request");
		exit (2);
	}
	r.r_uid = getuid();
	r.r_gid = getgid();
	strncpy (r.r_que, queue, QUESIZ);		/* Paranoid */
	if (puthdr (&r, rfp) != 0) {
		netclose ("406 Cannot rewrite request");
		exit (2);
	}
	fflush (rfp);
	if (ferror (rfp)) {
		netclose ("407 Error rewriting request");
		exit (2);
	}
	fclose (rfp);

	signal (SIGPIPE, SIG_IGN);
	pipe (pfd);
	switch (fork()) {
	case -1:
		netclose ("410 Error during fork");
		exit (2);
	case 0:
		/* Child */
		close (pfd[0]);
		dup2 (pfd[1], 1);
		close (pfd[1]);
		execl (Queue2, "queue2", file, "-v", 0);
		exit (99);
	default:
	    {
		char	message[256];
	    	char	*cp;
	    	int	nread;

		if ((nread = read (pfd[0], &message[4], (sizeof message) - 5)) > 1)
			message[nread+4] = 0;
	    	else
			strcpy (message, "....No message from queue2");
	    	if (cp = index(message, '\n'))
	    		*cp = 0;
		wait (&status);
		if (status&0377 != 0) {
			netclose ("411 Abnormal status from MDQS/Queue2");
			exit (2);
		}
		if ((status>>8) & 0377)
			strncpy(message, "411 ", 4);
		else
			strncpy(message, "250 ", 4);
		if (netclose (message))
			exit (0);
		else
			exit (2);
	    }
	}
}

netreply (fmt, a, b, c, d)
char	*fmt;
char	*a, *b, *c, *d;
{
	char	buf[256];
	int	len;
	
	if (!interactive)
		return;

if (debug) {
	fputs ("NETRECV: ->", stderr);
	fprintf (stderr, fmt, a, b, c, d);
	fputc ('\n', stderr);
}
	alarm (20);
	sprintf (buf, fmt, a, b, c, d);
	strcat (buf, "\n");
	len = strlen(buf);

	if (write (1, buf, len) < len) {
if (debug) fprintf (stderr, "NETRECV: Netwrite error\n");
		exit (2);
	}
	alarm (0);
	return (0);
}		

netclose (msg)
char	*msg;
{
	char	buf[256];
	int	code;
	int	len;
	
if (debug) fprintf (stderr, "NETRECV: %s\n", msg);
	code = atoi (msg);		/* For verification */
	sprintf (buf, "%s\n", msg);
	len = strlen (buf);

	alarm (20);
	if (write (1, buf, len) < len) {
if (debug) fprintf (stderr, "NETRECV: Netwrite error\n");
		exit (2);
	}
	alarm (0);
	if (interactive) {
		alarm (20);
		fgets (buf, 20, stdin);
		alarm (0);
		return (lexnequ(buf, "Bye ", 4) && atoi(&buf[4]) == code);
	} else {
		return (1);
	}
}		

alarmtrap ()
{
	fprintf (stderr, "NETRECV: Timeout site %s, request %s\n",
		site, request);
	exit (2);
}

/*
 *		E R R ( )
 *  Calls fprintf on the stderr and feeds in the format and one arg.
 */
/* VARARGS */
err (format, a, b, c, d)
char	*format;
char	*a, *b, *c, *d;
{
	fputs ("NETRECV: ", stderr);
	fprintf (stderr, format, a, b, c, d);
	putc('\n', stderr);
	fflush (stderr);
}
