#
/*
 *	    M U L T I - D E V I C E   Q U E U E I N G   S Y S T E M
 *
 *
 *			 Queueing Process 2 (QUEUE2)
 *
 *			   Douglas P. Kingston III
 *
 *		 Ballistics Research Laboratory,  June 1982
 *
 *	________________________________________________________________
 *
 *				Revision History
 *				================
 *	 6 Dec 82  DPK	Added xmalloc routine called by locking code.
 *
 *	11 Apr 83  DPK  Fixed declaration of args to complain().
 */

#include <stdio.h>
#include <signal.h>
#include "queue.h"
#include <sys/stat.h>
#ifdef SYS3
#include <varargs.h>
#endif

extern	char	*Qcntrldir;
extern	char	*Newprefix;
extern	char	*Tmpqdir;
extern	char	*Qdatadir;
extern	lock_t	Seq_lock;

char	*index();
struct request *gethdr();

struct	stat	statb;

/*
 *		M A I N ( )
 *  Argv[1] has the short form name of the queue control file
 *  (e.g. Qpr00451).  It is assumed to be in Tmpqdir.
 */
main (argc, argv)
int	argc;
char	**argv;
{
	int	seqno;
	int	uid;
	int	gid;
	int	verbose = 0;		/* Print queued name if non-zero */
	int	die = 0;		/* If non-zero, then panic exit */
	char	*cp;			/* For tripping thru linebuf */
	FILE	*cf;			/* The control file */
	struct	request r;		/* Request structure from cf */
	char	linebuf[PATHSIZ];	/* Lines from control file */
	char	tmpcname[PATHSIZ];	/* Initial control file name */
	char	Pcntrlname[QDIRSIZ];	/* Tmp control file name */
	char	Qcntrlname[QDIRSIZ];	/* Final control file name */
	char	tmpdname[PATHSIZ];
	char	newname[PATHSIZ];
	char	obuf[BUFSIZ];		/* For buffered STDOUT */
	
	signal (SIGHUP, SIG_IGN);
	signal (SIGINT, SIG_IGN);
	signal (SIGQUIT, SIG_IGN);
	umask( 077 );

	fdverify();			/* Paranoid */
	setbuf( stdout, obuf );

	if (argc < 2) {
		printf ("Usage:  %s tmpqfile [-v]\n");
		exit (1);
	}
	if (argc == 3 && strcmp(argv[2], "-v") == SAME)
		verbose++;

	/*
	 *	QUEUE2 is setuid to Daemon so the following chdir will work
	 */
	if (chdir (Qcntrldir) < 0) {
		complain( "Can't chdir to %s: %s\n",
			   Qcntrldir, sys_errlist[errno] );
		die++;
	}
	setgid (gid = getgid());
	setuid (uid = getuid());

	/*
	 *	Now unprivileged
	 */
	sprintf (tmpcname, "%s/%s", Tmpqdir, argv[1]);

	/*  Get a sequence number  */
	if( lock (&Seq_lock, 60) < 0 )  {	/****  Seq_lock LOCKED  ****/
		complain( "Can't lock sequence file (%s)", sys_errlist[errno]);
		fatal( 13 );
	}
	seqno = nextseqno (uid);
	unlock (&Seq_lock);
	if( seqno < 0 ) {
		complain( "Cannot get sequence number.");
		fatal( 14 );
	}
	sprintf (Pcntrlname, "P%05.5d.%d", uid, seqno);	/* Temporary name */
	sprintf (Qcntrlname, "Q%05.5d.%d", uid, seqno);	/* Final name */

	/*
	 *  SECURITY MEASURE!
	 *  Make sure we have the one and only copy of the file.
	 *  On USG Unix systems there is a special danger.  If
	 *  the user creates a link to the control file, he could
	 *  later overwrite it and then chown it to root or some
	 *  such.  This is defeated by making sure the link count
	 *  is 1 on the Pcntrlname file.
	 */
	if (link (tmpcname, Pcntrlname) < 0) {
		complain( "Can't link %s to %s: %s\n",
			  tmpcname, Pcntrlname, sys_errlist[errno] );
		fatal (11);
	}
	unlink (tmpcname);	/* No user can swap files now. */

	if(( cf = fopen( Pcntrlname, "r+" )) == NULL  ||
	     gethdr (cf, &r) == (struct request *) NULL ) {
		complain( "Can't get request %s: %s\n",
			  Pcntrlname, sys_errlist[errno] );
		fatal (12);
	}
	fstat (fileno(cf), &statb);
	if (statb.st_nlink > 1) {
		/* This check is CRITICAL on USG systems */
		/* Possible security breach attempt, be paranoid */
		complain( "Too many links to %s (%d)", Pcntrlname,
			  statb.st_nlink);
		fatal (13);
	}

	if (die) {
		rmdata( cf );
		unlink( Pcntrlname );
		fatal( 10 );
	}

	while (fgets(linebuf, sizeof (linebuf), cf)) {
		if (cp = index (linebuf, '\n'))
			*cp = 0;
		if (linebuf[0] == 'I' || linebuf[0] == 'B') {
			sprintf (tmpdname, "%s%s", Tmpqdir, &linebuf[1]);
			sprintf (newname, "%s%s", Qdatadir, &linebuf[1]);
			if (link (tmpdname, newname) < 0) {
				complain( "Cannot link in %s: %s\n",
					  tmpdname, sys_errlist[errno] );
				fatal (13);
			}
			unlink (tmpdname);
		}
	}

	if (fseek (cf, 0L, 0) == (-1L)) {
		complain( "Fseek error on %s: %s\n",
			   Pcntrlname, sys_errlist[errno] );
		fatal (12);
	}

	/*
	 *  Following values are the business of queue2 only, not the user
	 */
	strncpy (r.r_name, Qcntrlname, QDIRSIZ);
	r.r_uid = uid;
	r.r_gid = gid;
 	time (&r.r_tsubmit);
	if (r.r_tstart == 0L  || r.r_tstart < r.r_tsubmit)
		r.r_tstart = r.r_tsubmit;	/* No start before submit */

	puthdr (&r, cf);
	fflush (cf);
	if (ferror (cf)) {
		complain( "Could not rewrite %s: %s\n",
			  Pcntrlname, sys_errlist[errno] );
		fatal (12);
	}
	fclose( cf );

	/*
	 *	The last step is to link the control file to the proper
	 *	Q* name and then link it into the ../new directory so
	 *	that the deamon will spot it and read it in.
	 */
	strcpy( newname, Newprefix );
	strcat( newname, Qcntrlname );
	if( link( Pcntrlname, Qcntrlname ) < 0 ) {
		complain( "Can't link %s to %s: %s",
			  Pcntrlname, Qcntrlname, sys_errlist[errno] );
		fatal (12);
	}
	unlink (Pcntrlname);
	if ( link( Qcntrlname, newname ) < 0 ) {
		complain( "Can't link %s to %s: %s",
			  Qcntrlname, newname, sys_errlist[errno] );
		fatal (12);
	}

	if (verbose)
		printf ("Queued %s\n", r.r_name);
	proddaemon();
	exit (0);
}

/*
 *		C O M P L A I N ( )
 *  Print an error message for a less serious (non-permanent)
 *  error.  Used by the lock code.
 */
/* VARARGS */
#ifdef SYS3
complain( va_alist )
	va_dcl
{
	char	*fmt;
	va_list	ap;

	fprintf( stderr, "Queue2: " );
	va_start( ap );
	fmt = va_arg( ap, char * );
	vfprintf( stderr, fmt, ap );
	va_end( ap );
	putc( '\n', stderr );
}
#else	/* 4.2BSD doesn't have vfprintf; use usual kludge and hope it works */
complain( fmt, a, b, c, d )
char	*fmt, *a, *b, *c, *d;
{
	fprintf( stderr, "Queue2: " );
	fprintf( stderr, fmt, a, b, c, d );
	putc ('\n', stderr);
}
#endif

/*
 *		F A T A L ( )
 *  Die from horrible causes.  Print a message that
 *  will hopefully cause the user to seek experienced help.
 */
fatal (code)
int	code;
{
	fprintf (stderr, "Try again or seek help from systems staff\n");
	exit (code);
}
