#
/*
 *	    M U L T I - D E V I C E   Q U E U E I N G   S Y S T E M
 *
 *			   Douglas P. Kingston III
 *
 *		      Verset Typesetter Emulator Server
 *
 *	Note that a pipe to the Deamon will be on file descriptor 3.
 *	________________________________________________________________
 *
 *				Revision History
 *				================
 */

#include <stdio.h>
#include <signal.h>
#include <errno.h>
#include "queue.h"
#include "qreplys.h"

#define	ENORUN	79	/* Error return from Vers0 when /dev/vpt is busy */

struct	request r;
char	*vers0	= "/usr/bin/vers";
char	*output = "/dev/null";
int	pid;

/* 100 Verset variables ought to be enough ... */
#define NVERVARS	100
char	*vargs[NVERVARS] = { "vers0", 0 };
char	inputfile[PATHSIZ];		/* Name of datafile with cmds */

extern	char *Qdatadir;

char	*index();
char	*xmalloc();
char	*malloc();
int	dorestart();
int	doflush();
struct	request *gethdr();

/*
 *		M A I N  ( )
 *  As a MDQS queue server, this process will be started up
 *  with the queue control file as the standard input, and
 *  the device of interest on the standard output.
 */
main (argc, argv)
int	argc;
char	**argv;
{
	char	linebuf[LINESIZ];
	char	*workdir;
	char	*cp;
	int	fin;
	int	status;
	int	i;
	
	signal( SIGHUP, SIG_IGN );		/* Paranoid. */
	signal( SIGEMT, dorestart );
	signal( SIGTERM, doflush );

	if (gethdr (stdin, &r) == (struct request *) NULL) {
		fprintf (stderr, "VSERVER: Error reading request\n");
		retmsg (RP_FATAL, "Error reading request header");
		exit (10);
	}

	while (fgets (linebuf, sizeof (linebuf), stdin) != NULL) {
		if ((cp = index (linebuf, '\n')) != 0)
			*cp = 0;
		switch (linebuf[0]) {
		case 'A':
			setargs( &linebuf[1] );
			break;

		case 'D':
			workdir = xmalloc( strlen( linebuf ));
			strcpy( workdir, &linebuf[1] );
			break;

		case 'I':
			sprintf( inputfile, "%s/%s", Qdatadir, &linebuf[1] );
			fin = open( inputfile, READ );
			if( fin < 0 ) {
				fprintf( stderr, "VSERVER: Can't open command file '%s'\n", inputfile );
				retmsg (RP_FATAL, "Cannot open source file '%s'", inputfile);
				exit( 99 );
			}
			break;

		case 'P':
			output = xmalloc( strlen( linebuf ));
			strcpy( output, &linebuf[1] );
			break;

		case 'U':
		case 'T':
		default:
			/* Ignore unknown control lines */
			break;
		}
	}

	/*
	 *	Now we are read to start the Vers0 process.
	 *	We are already running with our UID/GID
	 *	the same as the invoker.
	 */
	close(1);	/* So Vers0 can open it... */

	if(( pid = fork()) == -1 )
		retmsg (RP_RESTART, "Unable to fork()");

	else if( pid == 0 ) {
		/* Child */
		if( chdir( workdir ) < 0 )
			chdir( "/tmp" );	/* Its a safe default */

		close( 0 );
		dup2( fin, 0 );			/* Standard input is cmds */
		if(( i = open( output, WRITE )) < 0 ) {
		    if( errno != ENOENT || (i = creat( output, 0644 )) < 0) {
fprintf( stderr, "VSERVER: couldn't open output '%s', request refused\n",
	output );
		    	retmsg (RP_FATAL, "Cannot open output file '%s'", output);
			exit( 99 );
		    }
		}
		dup2( i, 1 );
		dup2( i, 2 );
#ifndef _NFILE
		for( i = getdtablesize()-1; i > 2; i--)
#else
		for( i = _NFILE-1; i > 2; i--)
#endif
			close( i );

		execv( vers0, vargs );
fprintf( stderr, "VSERVER: exec of %s failed (%s)\n",
			vers0, sys_errlist[errno]);
		abort();
	} else {
		/* Parent */
		wait( &status );
		if( status & 0xff ) {
fprintf( stderr, "VSERVER: request %s, vers0 bombed\n", r.r_name );
			retmsg (RP_HOLD, "vers0 bombed");
			unlink( inputfile );
		} else if( ((status>>8) & 0xff) == ENORUN ) {
			/* Can't open the device */
			sleep( OPENWAIT );
			retmsg (RP_NOPEN, "Cannot open versatec");
		} else {
			retmsg (RP_OK, 0);
			unlink( inputfile );
		}
	}

	exit (0);
}

/*
 *		S E T A R G S ( )
 *	This function is responsible for building the argument
 *	variable vector.  Pass it a string to be installed.
 */
setargs( varstr )
register char	*varstr;
{
	static	int	nused = 1;	/* arg0 is already supplied above */

	if( nused < NVERVARS - 1 ) {
		vargs[nused] = xmalloc( strlen( varstr ) + 1 );
		strcpy( vargs[nused], varstr );
		nused++;
		vargs[nused] = 0;
	}
}

/*
 *		X M A L L O C ( )
 *	Just malloc(), but it always returns a valid pointer or exits
 */
char *
xmalloc( size )
int	size;
{
	register char	*cp;

	if(( cp = malloc( size )) == NULL ) {
		fprintf( stderr, "VSERVER: No free memory!\n");
		retmsg (RP_FATAL, "No free memory");
		if( *inputfile )
			unlink( inputfile );
		exit( 0 );
	}
	return( cp );
}

/*
 *	DORESTART ()
 *
 *  This function is called by a signal to restart a job.
 */
dorestart()
{
	retmsg (RP_RESTART, "RESTART requested");
	if( pid )
		kill( pid, SIGKILL );
	exit( 0 );
}


/*
 *	DOFLUSH ()
 *
 *  This function is called by a signal to "flush" the job.
 */
doflush()
{
	retmsg (RP_FLUSH, "Server received flush request");
	if( pid );
		kill( pid, SIGKILL );
	exit( 0 );
}
